#!/usr/bin/python

# talKZ - (Kalfa and Zack)'s talk implementation
# Copyright (C) 2002:
# 	Cosimo Alfarano      <kalfa@cs.unibo.it>
# 	Stefano Zacchiroli   <zack@cs.unibo.it>
# 
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


""" talKZ client """


import curses.ascii, getopt, os, sys
from select import select
import debugger, logger, ncursesUI, peer, protocol
from messages import CAPA


debug = False


class TalKZClientConf(object):
    """
    talKZ client configuration

    Methods:
        setCallee(user)
        getCallee()
        setCalleeTty(tty)
        getCalleeTty()
        setCalleeHost(host)
        getCalleeHost()
        setLog(bool)
        getLog()
        setLogfile(logfile)
        getLogfile()
        setTls(bool)
        getTls()
    """

    def __init__(self, callee=None, callee_host=None, callee_tty=None,
                log=False, tls=False):
        self.setCallee(callee)
        self.setCalleeHost(callee_host)
        self.setCalleeTty(callee_tty)
        self.setLog(log)
        self.setTls(tls)

    def setCallee(self, callee):
        self.__callee = callee

    def getCallee(self):
        return self.__callee

    def setCalleeTty(self, callee_tty):
        self.__callee_tty = callee_tty

    def getCalleeTty(self):
        return self.__callee_tty

    def setCalleeHost(self, callee_host):
        self.__callee_host = callee_host

    def getCalleeHost(self):
        return self.__callee_host

    def setLog(self, log):
        if log not in [True, False]:
            raise ValueError("'log' must be a boolean")
        self.__log = log

    def getLog(self):
        return self.__log

    def setLogfile(self, logfile):
        if type(logfile) != str:
            raise TypeError("'logfile' must be a string")
        self.__logfile = logfile

    def getLogfile(self):
        return self.__logfile

    def setTls(self, tls):
        if tls not in [True, False]:
            raise ValueError("'tls' must be a boolean")
        self.__tls = tls

    def getTls(self):
        return self.__tls


class TalKZClient(object):
    """ talKZ client, just instantiate this class from a main """

    def __init__(self, conf=TalKZClientConf()):
        assert __name__ == "__main__"
        if debug:
            self.__debugger =   \
                debugger.Debugger(fname="TALKZ.PY.LOG", append=True)
        self.__conf = conf
        self.__parseConfFile()
        self.__parseCmdLine()
        if self.__conf.getLog():    # logging enabled
            self.__logger = logger.Logger(logfile = self.__conf.getLogfile(),
                        lname = os.getlogin(), rname = self.__conf.getCallee())
            if debug:
                self.__debugger.log("Logging using: %s" % `self.__logger`)
        else:   # logging disabled
            self.__logger = None
            if debug:
                self.__debugger.log("Logging not enabled!")
        self.__ui = ncursesUI.NCursesUI(self.__logger)
#         notifier = (lambda msg: self.__ui.setStatusMsg(msg))
        notifier = (lambda msg: sys.stdout.write(msg + "\n"))
        myCErase, myWErase, myLErase = self.__ui.getMagicCharsL()
        if self.__conf.getTls():    # enable TLS
            capabilities = CAPA["tls"]
        else:   # don't enable TLS
            capabilities = CAPA["none"]
        # talk protocol mandate the order in which magic chars are sent:
        # 1st: erase_char, 2nd: erase_line, 3rd: erase_word
        conn = protocol.ClientConnection(
                    mymagicchars = (myCErase, myLErase, myWErase),
                    capa = capabilities,
                    ruser = self.__conf.getCallee(),
                    rhost = self.__conf.getCalleeHost(),
                    rtty = self.__conf.getCalleeTty())
        try:
            ((hisCErase, hisLErase, hisWErase), self.__peer, backlog) = \
                conn.connect(notifier)
        except protocol.Aborted, e:
            self.__confirmExit("Failure: %s" % e.getReason())
        assert isinstance(self.__peer, peer.Peer)
        self.__ui.setMagicCharsR((hisCErase, hisWErase, hisLErase))
        if debug:
            self.__debugger.log("his magics = <c=%d, l=%d, w=%d>" % \
                                (hisCErase, hisLErase, hisWErase))
        try:
            self.__ui.start()
            self.__ui.setStatusMsg("Talking ...")
            for c in backlog:
                self.__handlePeerChar(c)
            self.__mainloop()
        finally:
            self.__cleanUp
        sys.exit(0)

    def __parseConfFile(self):
        # TODO parse a configuration file and change self.__conf accordingly
        # not yet needed!
        pass

    def __usageExit():
        """ invocation error, display usage message and exit with return code 1
        """

        print "Usage:"
        print "     talkz.py [options] user[@host] [tty]"
        print "Options:"
        print "     -s | --ssl                  enable SSL"
        print "     -l <file> | --log=<file>    enable logging to file <file>"
        print "     -h | --help                 show this help message"
        sys.exit(1)
    __usageExit = staticmethod(__usageExit)

    def __cleanUp(self):
        if self.__ui.isRunning():
            self.__ui.stop()
        if self.__logger != None:
            self.__logger.stop()

    def __niceExit(self):
        """ nice exit, talk completed properly, exit with return code 0 """

        self.__cleanUp()
        sys.exit(0)

    def __confirmExit(self, msg):
        """ display 'msg' in status bar, wait for an user confirmation and then
        exit nicely; if UI isn't rennuning display 'msg' on standard output and
        exit without waiting for any user interaction """

        if type(msg) != str:
            raise TypeError("'msg' must be a string")
        if self.__ui.isRunning():   # UI up and running
            self.__ui.setStatusMsg(msg)
            try:
                while True:
                    self.__ui.getChar()
            except KeyboardInterrupt:
                self.__niceExit()
        else:   # UI not yet started
            print msg
            self.__niceExit()

    def __parseCmdLine(self):
        """ parse command line arguments and change self.__conf accordingly """

        try:
            optlist, args = getopt.getopt(sys.argv[1:],
                        "shl:", ["ssl", "help", "log="])
        except getopt.GetoptError:
            self.__usageExit()
        if len(args) >= 1:  # user was given
            try:    # user is like "user@host"
                (ruser, rhost) = args[0].split("@")
                self.__conf.setCallee(ruser)
                self.__conf.setCalleeHost(rhost)
            except ValueError:  # user is more like "user"
                self.__conf.setCallee(args[0])
        if len(args) == 2:  # also tty was given
            self.__conf.setCalleeTty(args[1])
        elif (len(args) == 0) or (len(args) > 2):   # wrong number of args
            self.__usageExit()
        try:    # parse switches
            i = iter(optlist)
            while True:
                opt, value = i.next()
                if opt in ["-l", "--log"]:  # enable logging
                    self.__conf.setLog(True)
                    self.__conf.setLogfile(value)
                elif opt in ["-s", "--ssl"]:   # enable TLS
                    self.__conf.setTls(True)
                elif opt in ["-h", "--help"]:   # exit with usage message
                    self.__usageExit()
                else:   # impossible
                    raise "Arrrghh! I forgot to handle this option!!!!"
        except StopIteration:
            pass

    def __handleConsoleChar(self, ch):
        if debug:
            self.__debugger.log("received char %d from STDIN" % ch)
        if ch == curses.ascii.FF:   # explicit refresh request
            self.__ui.refresh()
        else:   # normal char
            if ch > 255:    # some 'strange' key
                ch = curses.ascii.ascii(ch)
            self.__ui.putCharL(ch) # local echo
            self.__peer.send(chr(ch))   # send char to peer

    def __handlePeerChar(self, ch):
        if ch == "":    # peer closed connection
            self.__confirmExit(
                "Connection closed, hit CTRL-C to exit ...")
        else:   # char received
                # display ch to remote talk win
            if debug:
                self.__debugger.log(
                    "received char %d from PEER" % ord(ch))
            self.__ui.putCharR(ord(ch))

    def __mainloop(self):
        try:
            while True:
                ready_fd = (select([sys.stdin, self.__peer], [], []))[0]
                if sys.stdin in ready_fd:   # data from stdin
                    self.__handleConsoleChar(self.__ui.getChar())
                if self.__peer in ready_fd: # data from peer via socket
                    self.__handlePeerChar(self.__peer.recv(1))
        except KeyboardInterrupt:
            self.__niceExit()

def main():
    conf = TalKZClientConf()
    TalKZClient(conf)


if __name__ == '__main__':
    main()

